//****************************************************************************
// 6510/6502 verilog cpu. with asynchonous memory interface
//
// (C) 2013, Ken Chamberlain <kenneth.r.chamberlain@comcast.net>
//
// Feel free to use this code in any project (commercial or not), as long as you
// keep this message, and the copyright notice. This code is provided "as is", 
// without any warranties of any kind. 
// 
//****************************************************************************

//****************************************************************************
//X6510A

//The 32 bit read port is used for reading instructions, vector addresses
// or external operands in page 2 or higher.

//The 8 bit read port is used for io operand read operations.

//The 8 bit write port is used for all write operations. Stacking operations
//are not written to external memory.

//Most instructions take fewer clock cycles than the original processor.
//mos6502 clock cycle counts, X6510A cycle counts are in the iDecode file.

//CPU states (S_EXECUTE_INSTn) are the starting instruction execute states
//for all instructions based on if it is a 1, 2 or 3 byte instruction. Some
//previous state was used to load and decode the instruction.
//To calculate cycle counts start counting at state S_EXECUTE_INSTn until you
//get back to a state S_EXECUTE_INSTn. Do not count state S_EXECUTE_INSTn
//again as that is now processing the next instruction.

//Unimplemented instructions will halt the processor.

//BCD arithmatic is supported but takes 1 additional clock cycle to complete.

//The processor can be stopped at any time with a high level on pause or
//low levels on vic_req or dma_req.

//Defaut memory accesses are instruction fetches, if there is room for
//another 4 bytes in the 12 byte ir queue they will be loaded into the ir queue
//If the instruction queue cannot take in 4 new bytes the next instruction
//fetch address is still read but the data is disregarded.

//Two types of dma are supported, burst type which is used by VIC requests
//for video and single cycle requests that can be used by some other device.
//Burst requests are acknowledged by vic_ack going low and will always occur
//on the next memory cycle of the request. A single cycle dma request is
//acknowleged by dma_ack. This is a lower priority dma and may be held off
//by a burst request at the same time or already in progress. The burst
//requests can never write to memory currently.

//`define N6502  //remove the C64 memory map register...
//`define debug  //video display of registers, pc breakpoint & single stepping.

//              __    __    __    __    __    _     __    __    __    __    __    __    __    __
//     clk  |__|  |__|  |__|  |__|  |__|  |__| ..._|  |__|  |__|  |__|  |__|  |__|  |__|  |__|
//
// cpu_a(al)   X A0  X A1  X A2  X A3  X V1  X ... X Vm  X Vn  X XX  X A4  X A5  X A6  X A7  X
//
//  cpu_rd     X i0  X i1  X i2  X i3  X ... X XX  X XX  X XX  X XX  X i4  X i5  X i6  X
//
// cpu_aw(awl) X B0  X B1  X B2  X B3  X XX  X ... X XX  X XX  X D0  X B4  X B5  X B6  X B7  X
//             _______       __________________   ______________       _________________________
//  cpu_rw            |_____|                  ...              |_____|
//
//  cpu_db     X O0  X XX  X O2  X O3  X ... X XX  X XX  X XX  X XX  X O4  X O5  X O6  X
//
//  cpu_wb     X XX  X w1  X XX  X XX  X XX  X ... X XX  X XX  X wd  X XX  X XX  X XX  X XX  X
//             ____________________                         ____________________________________
// vic_req*                        |___________..._________|
//             _________________________                         _______________________________
// vic_ack*                             |______...______________|
//             ____________________________                            _________________________
// dma_req*                                |___...____________________|
//             ________________________________   ______________       _________________________
// dma_ack*                                    ...              |_____|
//                                       ______   ____________________
// cpu_ack*  XX_________________________|      ...                    |_________________________
//              __    __    __    __    __    _     __    __    __    __    __    __    __    __
//     clk  |__|  |__|  |__|  |__|  |__|  |__| ..._|  |__|  |__|  |__|  |__|  |__|  |__|  |__|
//                                        ^

module X6510A(clk,      // system memory clock
              cpu_a,    // [15:0] cpu/vic/dma address for accessing cpu_rd[31:16] data
              cpu_al,   // [15:2] cpu/vic/dma address for accessing cpu_rd[15:0] data
              cpu_rd,   // [31:0] instrution, vector or address fetching data bus
              pRead,    // indicates the access was a peripheral and should be read off of the cpu_db bus
              cpu_db,   // [7:0] data operand read data (IO read)
              cpu_wb,   // [7:0] data operand write data
              cpu_rw,   // read(H) / write(L) control signal for external accesses
              cpu_wbyte,// [3:0] write{L} byte selects (may not be needed depending on memory config) -
                        // only one of these will ever go low at a time.
              vic_req,  // burst read dma request
              vic_ack,  // burst read dma acknowledge - 1 cycle delay from request
              vic_a,    // [15:1] burst read source address => cpu_a, cpu_al
              dma_req,  // single cycle dma access request *must be released on the next clk after dma_ack is asserted
              dma_ack,  // single cycle dma acknowledge - next cycle free from burst request will be used
              dma_a,    // [15:0] single cycle dma source address => cpu_a, cpu_al, cpu_aw, cpu_awl
                        // currently access to the internal stack ind zp is not available - it should be made
                        // available through the write byte port. External ZP is kept in sync with internal memory though.
              dma_dw,   // [7:0] single cycle dma data write source => cpu_wb
              dma_rw,   // single cycle dma read(H) / write(L) control signal => cpu_rw
              nmi_req,  // negative edge non-maskable interrupt input
              irq,      // low level maskable interrupt input
              resetIn,  // low level cpu reset input
              cpu_pause // low level cpu pause input
             
`ifndef N6502
              , cpu_hiram  // C64 memory map control signal out
              , cpu_loram  // C64 memory map control signal out
              , cpu_charen // C64 memory map control signal out
`endif             
`ifdef debug
              , debug_add // [4:0] video select of what to be displayed on its current raster
              , debug_Dout// [47:0] 8 characters x 6 bits data value to be displayed - supports 64 different characters
              , debug_sel // [2:0] selects what dataset is to be displayed, reigisters, pc trace etc...
              , breakTrig // low level output that indicates the pc breakpoint was reached (should cause pause to occur)
              , break_pc  // [15:0] pc compare address for program breakpoint
              , cpu_ack   // this is low when the cpu has control of the memory bus - needed for single stepping
`endif
             );

input clk;

input vic_req;
input [15:1] vic_a;
input nmi_req, irq, resetIn, cpu_pause;
input [31:0] cpu_rd;
input [7:0] cpu_db;
input pRead;

output [15:0] cpu_a;
output [15:2] cpu_al;
output [7:0] cpu_wb;
output [3:0] cpu_wbyte;
output vic_ack, cpu_rw;

input dma_req, dma_rw;
input [15:0] dma_a;
input [7:0] dma_dw;
output dma_ack;

`ifdef debug
input [15:0] break_pc;
input [4:0] debug_add;
input [2:0] debug_sel;
output [47:0] debug_Dout;
output breakTrig;
output cpu_ack;
`endif

//acknowledge signals for bus mastership
reg vic_ack, dma_ack, cpu_ack;

`ifndef N6502
//****************************************************************************
output cpu_hiram, cpu_loram, cpu_charen;

//dma cannot write this register, it can write to the ram image though
assign cpu_w = ~(|cpu_a[15:1]) & cpu_a[0] & ~cpu_rw & ~cpu_ack;

//C64 cpu memory control register
reg [2:0] a1 = 3'h6;
assign cpu_loram = a1[0];
assign cpu_hiram = a1[1];
assign cpu_charen = a1[2];

always@(posedge clk)
  begin
    //update the cpu memory map control register
    if(~resetIn)   a1 <= 3'h6; //set it wrong just to know it changes
    else if(cpu_w) a1 <= cpu_wb[2:0];
  end
`endif             

//****************************************************************************
// parameter definitions
parameter V_NMI = 16'hFFFA;
parameter V_RST = 16'hFFFC;
parameter V_IRQ = 16'hFFFE;
parameter ZERO = 8'h00;
parameter ONE = 8'h01;
parameter NEG_ONE = 8'hFF;
parameter H = 1'b1;
parameter L = 1'b0;
parameter RD = 1'b1;
parameter WR = 1'b0;
parameter TRUE = 1'b1;
parameter FALSE = 1'b0;
parameter BYTEX = 4'hF;
parameter BYTE0 = 4'hE;
parameter BYTE1 = 4'hD;
parameter BYTE2 = 4'hB;
parameter BYTE3 = 4'h7;

//****************************************************************************
//cpu state machine state names

// put reset vector on address bus to read the startup pc
parameter S_RESET = 4'h0;
// load IR with first instruction bytes
parameter S_LOAD_INST = 4'h1;
// decode the first instruction of new code path
parameter S_DECODE_INST = 4'h2;
// execute 1 byte instruction
parameter S_EXECUTE_INST1 = 4'h3;
// execute 2 byte instruction
parameter S_EXECUTE_INST2 = 4'h4;
// execute 3 byte instruction
parameter S_EXECUTE_INST3 = 4'h5;
// indirect y operand base address load
parameter S_IND_Y = 4'h6;
// read memory write back operand
parameter S_RD_MEM_OP = 4'h7;
// write memory operand
parameter S_WR_MEM_OP = 4'h8;
// rts pc adjust (+1)
parameter S_REG_OP = 4'h9;
// give BCD arithmatic an extra cycle to complete
parameter S_BCD_DELAY = 4'hA;
// rts pc adjust (+1)
parameter S_RTS = 4'hB;
// process nmi
parameter S_NMI = 4'hC;
// process irq
parameter S_IRQ = 4'hD;
// reload instruction register - ir may be corrupted by self modifying code
parameter S_IR_W_HIT = 4'hE;
// load more instruction bytes
parameter S_IR_LOAD = 4'hF;

//state machine
reg [3:0] next_state, state = S_RESET;

//****************************************************************************
// State machine patterns...
//
// [***] Instruction sequences below end in going to one of the S_EXECUTE_INSn
// states. S_IR_W_HIT or potentionally S_IR_LOAD may occur before entering 
// S_EXECUTE_INSn under certain conditions (but not both).
//
// [INT] indicates execution of either the S_IRQ or the S_NMI exception state
// sequence and replaces the normal state and any subsequent states listed.
// The brk & rti instructions do not respond to the irq or nmi signals.
// The subsequent instruction will be executed before enabled interupts are
// processed. 
//
// [IR_HIT] sequence will only be entered if an instruction's terminating state
// is a write to a location that is in the ir queue. Interrupts override this
// requirement. This requirement overrides any need for a S_IR_LOAD state
// execution.
//
// [S_IR_LOAD -> ***] indicates a conditional extra state may be required to
// keep the instruction decoding process going. The next instrution to be
// executed must have a decode cycle done first with enough bytes for the
// entire instruction available in the instruction register or being read from
// memory before one of the S_EXECUTE_INSn states can be entered.
//
// default cpu_a  address XXX == {cpu_if, L} - next 4 bytes in instruction stream
// default cpu_aw address XXX == {ONE, spp1} - current value on the stack
//
// {cpu_if, L} accesses will only use the data and increment itself if
// there is room in the ir queue for 4 more bytes.
//
// # indicates a write cycle
//

//****************************************************************************
// EXCEPTIONAL EVENTS :
//
// reset :
//
// state  : S_RESET -> S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_RST>    <pc>           {cpu_if, L}
// cpu_aw : XXX        XXX            XXX

// irq :
//
// state  : S_IRQ ->   S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_IRQ>    <pc>           {cpu_if, L}
// cpu_aw : {ONE, sp}# {ONE, spp2}#   {ONE, spp1}#

// nmi :
//
// state  : S_NMI ->   S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_NMI>    <pc>           {cpu_if, L}
// cpu_aw : {ONE, sp}# {ONE, spp2}#   {ONE, spp1}#

// ir refill due to self modifying code :
//
// state  : S_IR_W_HIT -> S_DECODE_INST -> [***]
// cpu_a  : <pc>          {cpu_if, L}
// cpu_aw : XXX           XXX

//****************************************************************************
// ONE BYTE INSTRUCTIONS :
//
// brk :
//
// state  : S_EXECUTE_INST1 -> S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_IRQ>            <pc>           {cpu_if, L}
// cpu_aw : {ONE, sp}#         {ONE, spp2}#   {ONE, spp1}#

// rti :
//
// state  : S_EXECUTE_INST1 -> S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : XXX                <pc>           {cpu_if, L}
// cpu_aw : {ONE, spp1}        XXX            XXX

// rts :  in S_RTS the pc off the stack is incremented
//
// state  : S_EXECUTE_INST1 -> S_RTS -> [INT] [S_LOAD_INST -> S_DECODE_INST -> ***]
// cpu_a  : XXX                XXX             <pc>           {cpu_if, L}
// cpu_aw : {ONE, spp1}        XXX             XXX            XXX

// php, pha :
//
// state  : S_EXECUTE_INST1 -> [INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_aw : {ONE, sp}#

// pla : in S_REG_OP flags are updated
//
// state  : S_EXECUTE_INST1 -> S_REG_OP -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_aw : {ONE, spp1}        XXX

// plp :
//
// state  : S_EXECUTE_INST1 -> [INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_aw : {ONE, spp1}

// all other one byte instructions :
//
// state  : S_EXECUTE_INST1 -> [INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_aw : XXX

//****************************************************************************
// TWO BYTE INSTRUCTIONS :
//
// branch_taken :
//
// state  : S_EXECUTE_INST2 -> [INT] [S_DECODE_INST -> ***]
// cpu_a  : <branch>                  {cpu_if, L}
// cpu_aw : XXX                       XXX

// (_ix) instructions : write
//
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [INT] [IR_HIT] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#        
// cpu_aw : {ZERO,addx[7:0]}   <tmp_a>#      

// (_ix) instructions : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST2 -> S_RD_MEM_OP -> S_REG_OP -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>        {cpu_if, L} {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}   <tmp_a>        XXX         XXX

// (_ix) instructions : all other register operations
//
// state  : S_EXECUTE_INST2 -> S_RD_MEM_OP -> S_REG_OP -> [INT] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>        {cpu_if, L} 
// cpu_aw : {ZERO,addx[7:0]}   <tmp_a>        XXX

// (_iy) instructions : write - S_IND_Y used to add the y index
//
// state  : S_EXECUTE_INST2 -> S_IND_Y ->  S_WR_MEM_OP -> [INT] [IR_HIT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} <tmp_a>#
// cpu_aw : {ZERO,add[7:0]}    XXX         <tmp_a>#

// (_iy) instructions : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST2 -> S_IND_Y ->  S_RD_MEM_OP -> S_REG_OP -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} <tmp_a>        {cpu_if, L} {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}    XXX         <tmp_a>        XXX         XXX

// (_iy) instructions : all other register operations
//
// state  : S_EXECUTE_INST2 -> S_IND_Y ->  S_RD_MEM_OP -> S_REG_OP -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} <tmp_a>        {cpu_if, L} 
// cpu_aw : {ZERO,add[7:0]}    XXX         <tmp_a>        XXX

// sta_z,stx_z,sty_z : writes
//
// state  : S_EXECUTE_INST2 -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {ZERO,add[7:0]}#                   {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}#                   XXX

// _z : memory operations
//
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP ->   [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {cpu_if, L}        {ZERO,add[7:0]}#                 {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}    {ZERO,add[7:0]}#                 XXX

// _z : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST2 -> S_REG_OP -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}    XXX         XXX

// _z : all other register operations
//
// state  : S_EXECUTE_INST2 -> S_REG_OP -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_aw : {ZERO,add[7:0]}    XXX

// sta_zx,sty_zx : writes
//
// state  : S_EXECUTE_INST2 -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {ZERO,addx[7:0]}#                  {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}#                  XXX

// _zx : memory operations
//
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP ->    [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {cpu_if, L}        {ZERO,addx[7:0]}#                 {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}   {ZERO,addx[7:0]}#                 XXX

// _zx : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST2 -> S_REG_OP -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}   XXX         XXX

// _zx : all other register operations
//
// state  : S_EXECUTE_INST2 -> S_REG_OP -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_aw : {ZERO,addx[7:0]}   XXX

// stx_zy : write
//
// state  : S_EXECUTE_INST2 -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {ZERO,addy[7:0]}#                  {cpu_if, L}
// cpu_aw : {ZERO,addy[7:0]}#                  XXX

// ldx_zy : register load
//
// state  : S_EXECUTE_INST2 -> S_REG_OP -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_aw : {ZERO,addy[7:0]}   XXX

// adc,sbc immediate instructions in decimal mode :
//
// state  : S_EXECUTE_INST2 -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_aw : XXX                XXX

// all other immediate instructions and branches not taken:
//
// state  : S_EXECUTE_INST2 -> [INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_aw : XXX

//****************************************************************************
// THREE BYTE INSTRUCTIONS :
//
// jmp_i :
//
// state  : S_EXECUTE_INST3 -> [INT] S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <add>              <pc>           {cpu_if, L}
// cpu_aw : XXX                XXX            XXX

// jmp_a :
//
// state  : S_EXECUTE_INST3 -> [INT] S_DECODE_INST -> [***]
// cpu_a  : <add>              {cpu_if, L}
// cpu_aw : XXX                XXX

// jsr_a :
//
// state  : S_EXECUTE_INST3 -> S_DECODE_INST -> [INT] [***]
// cpu_a  : <add>              {cpu_if, L}
// cpu_aw : {ONE, spp2}#       {ONE, spp1}#

// sta_a,stx_a,sty_a : writes
//
// state  : S_EXECUTE_INST3 -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {add}#                             {cpu_if, L}
// cpu_aw : {add}#                             XXX

// _a : memory operations
//
// state  : S_EXECUTE_INST3 -> S_WR_MEM_OP -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {add}             <tmp_a>#                        {cpu_if, L}
// cpu_aw : {add}             <tmp_a>#                        XXX

// _a : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST3 -> S_REG_OP  -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {add}             {cpu_if, L}   {cpu_if, L}
// cpu_aw : {add}             XXX            XXX

// _a : all other register operations
//
// state  : S_EXECUTE_INST3 -> S_REG_OP -> [INT] [***]
// cpu_a  : {add}             {cpu_if, L}
// cpu_aw : {add}             XXX

// sta_ax : write
//
// state  : S_EXECUTE_INST3 -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {addx}#                            {cpu_if, L}
// cpu_aw : {addx}#                            XXX

// _ax : memory operations
//
// state  : S_EXECUTE_INST3 -> S_WR_MEM_OP -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {addx}             <tmp_a>#                       {cpu_if, L}
// cpu_aw : {addx}             <tmp_a>#                       XXX

// _ax : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST3 -> S_REG_OP  -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {addx}            {cpu_if, L}   {cpu_if, L}
// cpu_aw : {addx}            XXX           XXX

// _ax : all other register operations
//
// state  : S_EXECUTE_INST3 -> S_REG_OP -> [INT] [***]
// cpu_a  : {addx}             {cpu_if, L}
// cpu_aw : {addx}             XXX

// sta_ay : write
//
// state  : S_EXECUTE_INST3 -> [INT] [IR_HIT] [S_IR_LOAD -> ***] [***]
// cpu_a  : {addy}#                            {cpu_if, L}
// cpu_aw : {addy}#                            XXX

// _ay : adc,sbc in decimal mode
//
// state  : S_EXECUTE_INST3 -> S_REG_OP -> S_BCD_DELAY -> [INT] [***]
// cpu_a  : {addy}            {cpu_if, L}  {cpu_if, L}
// cpu_aw : {addy}            XXX           XXX

// _ay : all other register operations
//
// state  : S_EXECUTE_INST3 -> S_REG_OP -> [INT] [***]
// cpu_a  : {addy}             {cpu_if, L}
// cpu_aw : {addy}             XXX


//****************************************************************************
reg [7:0] ira, imm_op, ah;
reg next_N, next_Z, next_C, next_V;
//wire [7:0] irr;

//two_byte_a - this may not indicate a valid 2 byte instruction, it does
//indicate it is not a valid one or three byte instruction...

iDecode id(clk, ira, next_C, next_V, next_N, next_Z,
           two_byte, one_byte_a, two_byte_a, three_byte_a, branch_taken,
           asla, rola, lsra, rora, dex, dey, inx, iny,txa, tya, txs, tay,
           tax, tsx, pha, pla,clc, sec, cli, sei, cld, sed, clv, php, plp,
           ora_i, anda_i, eora_i, lda_i, cmpa_i,// adca_i, sbca_i,
           ora_z, anda_z, eora_z, lda_z, cmpa_z, bita_z,// adca_z, sbca_z,
           ora_zx, anda_zx, eora_zx, lda_zx, cmpa_zx,// adca_zx, sbca_zx,
           sta_ix, ora_ix, anda_ix, eora_ix, lda_ix, cmpa_ix,// adca_ix, sbca_ix,
           sta_iy, ora_iy, anda_iy, eora_iy, lda_iy, cmpa_iy,// adca_iy, sbca_iy,
           ora_a, anda_a, eora_a, lda_a, cmpa_a, bita_a,// adca_a, sbca_a,
           ora_ax, anda_ax, eora_ax, ldy_ax, lda_ax, cmpa_ax,// adca_ax, sbca_ax,
           ora_ay, anda_ay, eora_ay, lda_ay, ldx_ay, cmpa_ay,// adca_ay, sbca_ay,
           asl_z, rol_z, lsr_z, ror_z, dec_z, inc_z,
           asl_zx, rol_zx, lsr_zx, ror_zx, dec_zx, inc_zx,
           asl_a, rol_a, lsr_a, ror_a, dec_a, inc_a,
           asl_ax, rol_ax, lsr_ax, ror_ax, dec_ax, inc_ax,
           stx_z, ldx_z, cpx_z, stx_zy, ldx_zy, stx_a, ldx_a, cpx_a, ldx_i, cpx_i,
           sty_z, ldy_z, cpy_z, sty_zx, ldy_zx, sty_a, ldy_a, cpy_a, ldy_i, cpy_i,
           //bpl, bmi, bvc, bvs, bcc, bcs, bne, beq,
           brk, rti, rts, jmp_a, jmp_i, jsr_a,
           sta_z, sta_zx, //sta_a, sta_ax, sta_ay,
           ix, iy, st_zxy, st_axy, zx, zy, ax, ay, mem_op, reg_op, addsub, imm_addsub);//,
//           branchi, Bra, imm_a, imm_x, imm_y, imm_cmp,
//           flags, reg_a, reg_x, reg_y);
//           irr, nop);

//decoded instructions saved for use out of the S_EXECUTE_INSTn states
reg sta_iy_2, addsub_2, mem_op_2, jsr_a_2;//, sty, stx;
reg asl, rol, ror, lsr, inc, dec;
reg ora, bita, anda, eora, ldx, lda, ldy, cpx_2, cmpa_2, cpy_2;

//data and index registers
reg [7:0] a, x, y, alu_op;

//status register bits
reg negative, zero, carry, overflow, interrupt, decimal;

reg [15:0] next_pc;
reg [15:1] next_if;
reg [15:0] addx, addy;
reg [7:0] next_sp, next_a, next_y, next_x;

wire [7:0] next_spp1, next_spp2, next_spp3;
assign next_spp1 = next_sp + H;
assign next_spp2 = next_sp + 2'h2;
assign next_spp3 = next_sp + 2'h3;

//instruction registers
//bit 0 never used...accesses always on word boundaries (for 4 usable bytes)
reg [15:1] cpu_if;
wire [15:1] cpu_ifp4 = cpu_if + 15'h0002;

reg [95:0] ir;        //12 bytes of instruction queue
reg [3:0] iCount;     //how many instruction bytes currently in the ir register

//ifCycle (Ready) data buses
wire [7:0] ib0, ib1, ib2, ib3;
wire [31:0] ib32;
assign ib0 = cpu_if[1] ? cpu_rd[23:16] : cpu_rd[7:0];
assign ib1 = cpu_if[1] ? cpu_rd[31:24] : cpu_rd[15:8];
assign ib2 = cpu_if[1] ? cpu_rd[7:0]   : cpu_rd[23:16];
assign ib3 = cpu_if[1] ? cpu_rd[15:8]  : cpu_rd[31:24];
assign ib32 = { ib3, ib2, ib1, ib0 };

//for making sure we can decode the instruction we think that we have
assign four_valid = |iCount[3:2];
assign two_valid = four_valid | iCount[1];
assign one_valid = two_valid | iCount[0];
assign three_valid = four_valid | (&iCount[1:0]);
assign five_valid = iCount[3] | (iCount[2] & (|iCount[1:0]));
assign six_valid = iCount[3] | (&iCount[2:1]);
//assign seven_valid = iCount[3] | (&iCount[2:0]);

//if low an add is occurring else it is a sub or cmp
reg mode;

//alu outputs
wire [7:0] result, asl_op, rol_op, lsr_op, ror_op, dec_op, inc_op, or_op, and_op, eor_op, bcdResult;

//binary alu
accumulatorPlus aA(mode, a, alu_op, carry, result, Cout_a, is_zero_a, is_negative_a, is_overflow_a,
                   alu_op_0, alu_op_1, alu_op_255, alu_op_dec_neg, alu_op_inc_neg,
                   asl_op, rol_op, lsr_op, ror_op, dec_op, inc_op,
                   or_op, and_op, eor_op, or_op_zero, and_op_zero, eor_op_zero);
                   
//decimal mode alu
bcdAccumulator aBCD(mode, a, alu_op, carry, bcdResult, bcdCout_a, bcdZero_a, bcdNegative_a, bcdOverflow_a);

//compare alu
compare cA(a, alu_op, Cout_ca, is_zero_ca, N_ca);
compare cX(x, alu_op, Cout_x, is_zero_x, N_x);
compare cY(y, alu_op, Cout_y, is_zero_y, N_y);

//synchronizing interrupt registers
reg rirq = H;       //synchonize irq
reg rnmi = H;       //synchonize nmi
reg rnmi2 = H;      //this is used to determine a falling rnmi edge
reg nmi = H;        //this is used to save the falling edge occurance until the next instruction

//internally held interrupt vector pc caches
//beware if your system bank switches these values - should work for C64.
//if bit 17 is zero they are disabled until initialized by an actual interrupt event
//reg [16:0] irq_rom = 17'h00000;
//reg [16:0] irq_ram = 17'h00000;
//reg [16:0] nmi_rom = 17'h00000;
//reg [16:0] nmi_ram = 17'h00000;

//an unimplemented instruction or some other fatal event causes at least
//1 bit in this variable to be set low and stops the processor in a halt condition
reg [3:0] halt = 4'hF;

//write byte data register
reg [7:0] cpu_wb;
reg cpu_rw, cpu_rwz;

//byte select write control register
reg [3:0] cpu_wbyte, cpu_wbytez;

//address registers
reg [15:0] pc, pc_p1r, pc_p2r, pc_p3r, pc_p4r, pc_p5r;
reg [15:0] branch, tmp_a, cpu_a, cpu_aw;

//reading an address from a memory
wire [15:0] cpu_dar0, cpu_dar1, cpu_dar2, cpu_dar3;
assign cpu_dar0 = cpu_rd[15:0];
assign cpu_dar1 = cpu_rd[23:8];
assign cpu_dar2 = cpu_rd[31:16];
assign cpu_dar3 = { cpu_rd[7:0], cpu_rd[31:24] };

//operand address
wire [15:0] add;
assign add = ir[23:8];

//program counter increment options
wire [15:0] pcp1, pcp2, pcp3, pcp4, pcp5;
assign pcp1 = next_pc + 16'h0001;
assign pcp2 = next_pc + 16'h0002;
assign pcp3 = next_pc + 16'h0003;
assign pcp4 = next_pc + 16'h0004;
assign pcp5 = next_pc + 16'h0005;

//stack / zero page cache
assign cacherw_hit = ~(|cpu_aw[15:9]);

wire [7:2] cpu_awl;
wire [31:0] cache, caches, cachez;

ram256 zerop(clk, cpu_wbytez, ~cpu_aw[8] & cacherw_hit, cpu_aw[7:2], cpu_awl, { cpu_wb, cpu_wb, cpu_wb, cpu_wb }, cachez);
ram256 stack(clk, cpu_wbytez, cpu_aw[8] & cacherw_hit, cpu_aw[7:2], cpu_awl, { cpu_wb, cpu_wb, cpu_wb, cpu_wb }, caches);

assign cache = cpu_aw[8] ? caches : cachez;

wire [15:0] cpu_sar0, cpu_sar1, cpu_sar2, cpu_sar3;
assign cpu_sar0 = caches[15:0];
assign cpu_sar1 = caches[23:8];
assign cpu_sar2 = caches[31:16];
assign cpu_sar3 = { caches[7:0], caches[31:24] };

wire [15:0] cpu_zar0, cpu_zar1, cpu_zar2, cpu_zar3;
assign cpu_zar0 = cachez[15:0];
assign cpu_zar1 = cachez[23:8];
assign cpu_zar2 = cachez[31:16];
assign cpu_zar3 = { cachez[7:0], cachez[31:24] };

//stack pointers used for address bus
reg [7:0] sp, spp1, spp2, spp3;
//stack pointer variations
wire [7:0] spm1, spm2, spm3;

assign spm1 = sp - H;
assign spm2 = sp - 2'h2;
assign spm3 = sp - 2'h3;

//processor status stacking options
wire [7:0] sp_int, sp_brk;
assign sp_int = { negative, overflow, H, L, decimal, interrupt, zero, carry };
assign sp_brk = { negative, overflow, H, H, decimal, interrupt, zero, carry };

reg [16:0] stackit;

`ifdef debug

//debug trace registers
reg [959:0] pc_trace = 960'd0;
reg [959:0] ad_trace = 960'd0;
reg [959:0] aw_trace = 960'd0;

//for pausing cpu due to certain events - starts single step
reg breakTrig = FALSE;
reg breakTrig1 = FALSE;

reg [31:0] cpu_drr, cache_rr;
reg [15:0] cpu_a2_i, cpu_aw2_i;
reg [15:2] cpu_al_i;
reg [7:0] cpu_pr_r, cpu_wb_r;
reg [3:0] cpu_byte_enable, cpu_byte_enablez;
reg cpu_rw_r, cpu_rwz_r;
 
//for counting events
reg [31:0] cycleCount, cycleCount2;
reg enableCycleCount2;

`endif

//asynchronous control state machines
//****************************************************************************
//****************************************************************************
//ir decode
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_INST:
            begin
              case(pc[1:0])
                2'h0 :
                  begin
                    ira    = cpu_rd[7:0];
                    imm_op = cpu_rd[15:8];
                    ah     = cpu_rd[23:16];
                  end
                2'h1 :
                  begin
                    ira    = cpu_rd[15:8];
                    imm_op = cpu_rd[23:16];
                    ah     = cpu_rd[31:24];
                  end
                2'h2 :
                  begin
                    ira    = cpu_rd[23:16];
                    imm_op = cpu_rd[31:24];
                    ah     = cpu_rd[7:0];
                  end
                2'h3 :
                  begin
                    ira    = cpu_rd[31:24];
                    imm_op = cpu_rd[7:0];
                    ah     = cpu_rd[15:8];
                  end
              endcase
            end
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(four_valid)
               begin
                ira = ir[15:8];
                imm_op = ir[23:16];
                ah = ir[31:24];
               end
             else if(three_valid)
               begin
                 ira = ir[15:8];
                 imm_op = ir[23:16];
                 ah = ib0;
               end
             else if(two_valid)
               begin
                 ira = ir[15:8];
                 imm_op = ib0;
                 ah = ib1;
               end
             else
               begin
                 ira = ib0;
                 imm_op = ib1;
                 ah = ib2;
               end
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(five_valid)
                begin
                  ira = ir[23:16];
                  imm_op = ir[31:24];
                  ah = ir[39:32];
                end
              else if(four_valid)
                begin
                  ira = ir[23:16];
                  imm_op = ir[31:24];
                  ah = ib0;
                end
              else if(three_valid)
                begin
                  ira = ir[23:16];
                  imm_op = ib0;
                  ah = ib1;
                end
              else
                begin
                 ira = ib0;
                 imm_op = ib1;
                 ah = ib2;
                end
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
//              if(six_valid)
//                begin
                  ira = ir[31:24];
                  imm_op = ir[39:32];
                  ah = ir[47:40];
//                end
////these result in some invalid settings but are not necesary because of inserted S_IR_LOAD states 
//              else if(five_valid) //extra S_IR_LOAD will be forced
//                begin
//                  ira = ir[31:24];
//                  imm_op = ir[39:32];
//                  ah = ib0;
//                end
//              else if(four_valid) //extra S_IR_LOAD will be forced
//                begin
//                  ira = ir[31:24];
//                  imm_op = ib0;
//                  ah = ib1;
//                end
//              else //this should never happen??? extra S_IR_LOAD will be forced
//                begin
//                  ira = ib0;
//                  imm_op = ib1;
//                  ah = ib2;
//                end
            end
//****************************************************************************
          default:
            begin
              if(three_valid)
                begin
                  ira = ir[7:0];
                  imm_op = ir[15:8];
                  ah = ir[23:16];
                end
              else if(two_valid)
                begin
                  ira = ir[7:0];
                  imm_op = ir[15:8];
                  ah = ib0;
                end
              else if(one_valid)
                begin
                  ira = ir[7:0];
                  imm_op = ib0;
                  ah = ib1;
                end
              else
                begin
                  ira = ib0;
                  imm_op = ib1;
                  ah = ib2;
                end
            end
        endcase
      end   
    else
      begin
        ira = ir[7:0];
        imm_op = ir[15:8];
        ah = ir[23:16];
      end
  end

//****************************************************************************
//****************************************************************************
//write bus
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_NMI,S_IRQ:             cpu_wb = pc[15:8];
//****************************************************************************
          S_LOAD_INST:             cpu_wb = stackit[15:8];
//****************************************************************************
          S_DECODE_INST:           cpu_wb = stackit[7:0];
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk|php)           cpu_wb = pc[15:8];
              else                  cpu_wb = a;
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(stx_z|stx_zy)      cpu_wb = x;
              else if(sty_z|sty_zx) cpu_wb = y;
              else                  cpu_wb = a;
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a)             cpu_wb = pc_p2r[15:8];
              else if(stx_a)        cpu_wb = x;
              else if(sty_a)        cpu_wb = y;
              else                  cpu_wb = a;
            end
//****************************************************************************
          S_WR_MEM_OP: // complete memory operation
            begin
              if(asl)               cpu_wb = asl_op;
              else if(rol)          cpu_wb = rol_op;
              else if(lsr)          cpu_wb = lsr_op;
              else if(ror)          cpu_wb = ror_op;
              else if(dec)          cpu_wb = dec_op;
              else if(inc)          cpu_wb = inc_op;
              else                  cpu_wb = a;
            end
//****************************************************************************
          default:                  cpu_wb = a;
        endcase
      end
    else                            cpu_wb = dma_dw;
  end


//****************************************************************************
//****************************************************************************
//r/w control (of the write bus)
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(st_zxy) cpu_rw = WR;
              else       cpu_rw = RD;
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(st_axy) cpu_rw = WR;
              else       cpu_rw = RD;
            end
//****************************************************************************
          S_WR_MEM_OP:   cpu_rw = WR;
//****************************************************************************
          default:       cpu_rw = RD;
        endcase
      end
    else if(~dma_ack)    cpu_rw = dma_rw;
    else                 cpu_rw = RD;
  end


//****************************************************************************
//****************************************************************************
//read instruction, pc vector address bus
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_RESET:                       cpu_a = V_RST;
//****************************************************************************
          S_LOAD_INST:                   cpu_a = pc;
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk)                    cpu_a = V_IRQ;
              else                       cpu_a = { cpu_if, L };
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(branch_taken)           cpu_a = branch;
              else if(sta_z|stx_z|sty_z) cpu_a = { ZERO, add[7:0] };
              else if(sta_zx|sty_zx)     cpu_a = { ZERO, addx[7:0] };
              else if(stx_zy)            cpu_a = { ZERO, addy[7:0] };
              else                       cpu_a = { cpu_if, L };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(ax)                     cpu_a = addx;
              else if(ay)                cpu_a = addy;
              else                       cpu_a = add;
            end
//****************************************************************************
          S_RD_MEM_OP,S_WR_MEM_OP:       cpu_a = tmp_a;
//****************************************************************************
          S_NMI:                         cpu_a = V_NMI;
//****************************************************************************
          S_IRQ:                         cpu_a = V_IRQ;
//****************************************************************************
          S_IR_W_HIT:                    cpu_a = pc;
//****************************************************************************
          default:                       cpu_a = { cpu_if, L };
        endcase
      end
    else if(~dma_ack)                    cpu_a = dma_a;
    else                                 cpu_a = { vic_a, L };
  end

//lower 16 bit data bus address generation 
assign cpu_al = cpu_a[15:2] + cpu_a[1];

//****************************************************************************
//****************************************************************************
//write byte select
always@(*)
  begin
    if(~cpu_rw)
      begin
        case(cpu_a[1:0])
          2'h0 : cpu_wbyte = BYTE0;
          2'h1 : cpu_wbyte = BYTE1;
          2'h2 : cpu_wbyte = BYTE2;
          2'h3 : cpu_wbyte = BYTE3;
        endcase
      end
    else         cpu_wbyte = BYTEX;
  end

//****************************************************************************
//****************************************************************************
//separate r/w control of the internal memory
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_INST,S_DECODE_INST: cpu_rwz = ~stackit[16];
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk|php|pha)        cpu_rwz = WR;
              else                   cpu_rwz = RD;
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(st_zxy)             cpu_rwz = WR;
              else                   cpu_rwz = RD;
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a|st_axy)       cpu_rwz = WR;
              else                   cpu_rwz = RD;
            end
//****************************************************************************
          S_WR_MEM_OP,S_NMI,S_IRQ:   cpu_rwz = WR;
//****************************************************************************
          default:                   cpu_rwz = RD;
        endcase
      end
    else if(~dma_ack)                cpu_rwz = dma_rw;
    else                             cpu_rwz = RD;
  end


//****************************************************************************
//****************************************************************************
// write or read instruction operand address bus
always@(*)
  begin
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_LOAD_INST:             cpu_aw = { ONE, spp2 };
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk|pha|php)      cpu_aw = { ONE, sp };
              else                 cpu_aw = { ONE, spp1 };
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(zx|ix)            cpu_aw = { ZERO, addx[7:0] };
              else if(zy)          cpu_aw = { ZERO, addy[7:0] };
              else                 cpu_aw = { ZERO, add[7:0] };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a)            cpu_aw = { ONE, sp };
              else if(ax)          cpu_aw = addx;
              else if(ay)          cpu_aw = addy;
              else                 cpu_aw = add;
            end
//****************************************************************************
          S_WR_MEM_OP,S_WR_MEM_OP: cpu_aw = tmp_a;
//****************************************************************************
          S_NMI,S_IRQ:             cpu_aw = { ONE, sp };
//****************************************************************************
          default:                 cpu_aw = { ONE, spp1 };
        endcase
      end
    else                           cpu_aw = dma_a;
  end

//lower 16 bit cache data bus address generation 
assign cpu_awl = cpu_aw[7:2] + cpu_aw[1];

//****************************************************************************
//****************************************************************************
//write byte select
always@(*)
  begin
    if(~cpu_rwz)
      begin
        case(cpu_aw[1:0])
          2'h0 : cpu_wbytez = BYTE0;
          2'h1 : cpu_wbytez = BYTE1;
          2'h2 : cpu_wbytez = BYTE2;
          2'h3 : cpu_wbytez = BYTE3;
        endcase
      end
    else         cpu_wbytez = BYTEX;
  end

////****************************************************************************
////****************************************************************************
////next a look-ahead
//always@(*)
//  begin
//    case(state)
////****************************************************************************
////start executing a single byte instruction
//      S_EXECUTE_INST1:
//        begin
//          if(asla)                       next_a = { a[6:0], L };
//          else if(rola)                  next_a = { a[6:0], carry };
//          else if(lsra)                  next_a = { L, a[7:1] };
//          else if(rora)                  next_a = { carry, a[7:1] };
//          else if(tya)                   next_a = y;
//          else if(txa)                   next_a = x;
//          else                           next_a = a;
//        end
////****************************************************************************
////start executing a 2 byte instruction
//      S_EXECUTE_INST2:
//        begin
//          if(anda_i)                     next_a = and_op;
//          else if(eora_i)                next_a = eor_op;
//          else if(ora_i)                 next_a = or_op;
//          else if(lda_i)                 next_a = alu_op;
//          else if(imm_addsub & ~decimal) next_a = result;
//          else                           next_a = a;
//        end
////****************************************************************************
//      S_REG_OP: // complete register operation
//        begin
//          if(ora)                        next_a = or_op;
//          else if(anda)                  next_a = and_op;
//          else if(eora)                  next_a = eor_op;
//          else if(lda)                   next_a = alu_op;
//          else if(addsub_2 & ~decimal)   next_a = result;
//          else                           next_a = a;
//        end
////****************************************************************************
//      S_BCD_DELAY:                       next_a = bcdResult;
////****************************************************************************
//      default:                           next_a = a;
//    endcase
//  end

//****************************************************************************
//****************************************************************************
//next x look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(tax)      next_x = a;
          else if(tsx) next_x = sp;
          else if(dex) next_x = x - ONE;
          else if(inx) next_x = x + ONE;
          else         next_x = x;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(ldx_i)    next_x = ir[15:8];
          else         next_x = x;
        end
//****************************************************************************
      S_REG_OP:
        if(ldx)        next_x = alu_op;
        else           next_x = x;
//****************************************************************************
      default:         next_x = x;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next y look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(tay)      next_y = a;
          else if(dey) next_y = y - ONE;
          else if(iny) next_y = y + ONE;
          else         next_y = y;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(ldy_i)    next_y = ir[15:8];
          else         next_y = y;
        end
//****************************************************************************
      S_REG_OP:
        if(ldy)        next_y = alu_op;
        else           next_y = y;
//****************************************************************************
      default:         next_y = y;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next state look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_RESET,S_RTS,S_NMI,S_IRQ:                    next_state = S_LOAD_INST;  
//****************************************************************************
      S_LOAD_INST,S_IR_W_HIT:                       next_state = S_DECODE_INST;
//****************************************************************************
      S_DECODE_INST:
        if(jsr_a_2 & ~nmi)                          next_state = S_NMI;
        else if(jsr_a_2 & ~rirq)                    next_state = S_IRQ;
        else if(one_byte_a)                         next_state = S_EXECUTE_INST1;
        else if(three_byte_a)                       next_state = S_EXECUTE_INST3;
        else                                        next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST1:
        if(brk|rti)                                 next_state = S_LOAD_INST;
        else if(rts)                                next_state = S_RTS;
        else if(pla)                                next_state = S_REG_OP;
        else if(~nmi)                               next_state = S_NMI;
        else if(~rirq)                              next_state = S_IRQ;
        else if(one_byte_a)                         next_state = S_EXECUTE_INST1;
        else if(three_byte_a)                       next_state = S_EXECUTE_INST3;
        else                                        next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST2:
        if(sta_ix|mem_op)                           next_state = S_WR_MEM_OP;
        else if(ix)                                 next_state = S_RD_MEM_OP;
        else if(iy)                                 next_state = S_IND_Y;
        else if(imm_addsub & decimal)               next_state = S_BCD_DELAY;
        else if(reg_op)                             next_state = S_REG_OP;
        else if(~nmi)                               next_state = S_NMI;
        else if(~rirq)                              next_state = S_IRQ;
        else if(branch_taken)                       next_state = S_DECODE_INST;
        else if(((cpu_aw - pc_p2r) < 12) & ~cpu_rw) next_state = S_IR_W_HIT;
        else if(st_zxy)
          begin
            if(one_byte_a & three_valid)            next_state = S_EXECUTE_INST1;
            else if(three_byte_a & five_valid)      next_state = S_EXECUTE_INST3;
            else if(four_valid & two_byte_a)        next_state = S_EXECUTE_INST2;
            else                                    next_state = S_IR_LOAD;
          end
        else if(one_byte_a)                         next_state = S_EXECUTE_INST1;
        else if(three_byte_a)                       next_state = S_EXECUTE_INST3;
        else                                        next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST3:
        if(jsr_a)                                   next_state = S_DECODE_INST; 
        else if(mem_op)                             next_state = S_WR_MEM_OP;
        else if(reg_op)                             next_state = S_REG_OP;
        else if(~nmi)                               next_state = S_NMI;
        else if(~rirq)                              next_state = S_IRQ;
        else if(jmp_a)                              next_state = S_DECODE_INST; 
        else if(jmp_i)                              next_state = S_LOAD_INST; 
        else if(((cpu_aw - pc_p3r) < 12) & ~cpu_rw) next_state = S_IR_W_HIT;
        else if(one_byte_a & four_valid)            next_state = S_EXECUTE_INST1;
        else if(three_byte_a & six_valid)           next_state = S_EXECUTE_INST3;
        else if(five_valid & two_byte_a)            next_state = S_EXECUTE_INST2;
        else                                        next_state = S_IR_LOAD;
//****************************************************************************
      S_RD_MEM_OP: // reading memory operand
        if(mem_op_2)                                next_state = S_WR_MEM_OP; 
        else                                        next_state = S_REG_OP; 
//****************************************************************************
      S_IND_Y:  // indirect y
        if(sta_iy_2)                                next_state = S_WR_MEM_OP;
        else                                        next_state = S_RD_MEM_OP;
//****************************************************************************
      S_WR_MEM_OP:
        if(~nmi)                                    next_state = S_NMI;
        else if(~rirq)                              next_state = S_IRQ;
        else if(((cpu_aw - pc) < 12) & ~cpu_rw)     next_state = S_IR_W_HIT;
        else if(one_byte_a & one_valid)             next_state = S_EXECUTE_INST1;
        else if(three_byte_a & three_valid)         next_state = S_EXECUTE_INST3;
        else if(two_valid & two_byte_a)             next_state = S_EXECUTE_INST2;
        else                                        next_state = S_IR_LOAD;
//****************************************************************************
      S_REG_OP:
        if(addsub_2 & decimal)                      next_state = S_BCD_DELAY;
        else if(~nmi)                               next_state = S_NMI;
        else if(~rirq)                              next_state = S_IRQ;
        else if(one_byte_a)                         next_state = S_EXECUTE_INST1;
        else if(three_byte_a)                       next_state = S_EXECUTE_INST3;
        else                                        next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_BCD_DELAY,S_IR_LOAD:
        if(~nmi)                                    next_state = S_NMI;
        else if(~rirq)                              next_state = S_IRQ;
        else if(one_byte_a)                         next_state = S_EXECUTE_INST1;
        else if(three_byte_a)                       next_state = S_EXECUTE_INST3;
        else                                        next_state = S_EXECUTE_INST2;
//****************************************************************************
      default:                                      next_state = state; //should never get here
    endcase
  end

//****************************************************************************
//****************************************************************************
//next pc look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_RESET,S_NMI,S_IRQ:  //read the new program counter off the data bus
        begin
          case(cpu_a[1:0])
            0 :            next_pc = cpu_dar0;
            1 :            next_pc = cpu_dar1;
            2 :            next_pc = cpu_dar2;
            3 :            next_pc = cpu_dar3;
          endcase
        end
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(rts)
            case(cpu_aw[1:0])
              0 :          next_pc = cpu_sar0;
              1 :          next_pc = cpu_sar1;
              2 :          next_pc = cpu_sar2;
              3 :          next_pc = cpu_sar3;
            endcase
          else if(rti)
            case(cpu_aw[1:0])
              0:           next_pc = cpu_sar1;
              1:           next_pc = cpu_sar2;
              2:           next_pc = cpu_sar3;
              3:           next_pc = cpu_sar0;
            endcase
          else             next_pc = pc_p1r;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(branch_taken) next_pc = branch;
          else             next_pc = pc_p2r;
       end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a|jmp_a)  next_pc = add;
          else if(jmp_i)
            case(cpu_a[1:0])
              0 :          next_pc = cpu_dar0;
              1 :          next_pc = cpu_dar1;
              2 :          next_pc = cpu_dar2;
              3 :          next_pc = cpu_dar3;
            endcase
          else             next_pc = pc_p3r;
        end
//****************************************************************************
      S_RTS:               next_pc = pc_p1r;
//****************************************************************************
      default:             next_pc = pc;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next if look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_RESET,S_NMI,S_IRQ:
        begin
          case(cpu_a[1:0])
            0 :                  next_if = cpu_dar0[15:1] + 15'h0002;
            1 :                  next_if = cpu_dar1[15:1] + 15'h0002;
            2 :                  next_if = cpu_dar2[15:1] + 15'h0002;
            3 :                  next_if = cpu_dar3[15:1] + 15'h0002;
          endcase
        end
//****************************************************************************
      S_DECODE_INST,S_IR_LOAD:   next_if = cpu_ifp4;
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(rti)
            case(cpu_aw[1:0])
              0:                 next_if = cpu_sar1[15:1] + 15'h0002;
              1:                 next_if = cpu_sar2[15:1] + 15'h0002;
              2:                 next_if = cpu_sar3[15:1] + 15'h0002;
              3:                 next_if = cpu_sar0[15:1] + 15'h0002;
            endcase
          else if(iCount < 4'hA) next_if = cpu_ifp4;
          else                   next_if = cpu_if;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(branch_taken)     next_if = branch[15:1] + 15'h0002;
          else if(~st_zxy)
            if(iCount < 4'hB)  next_if = cpu_ifp4;
            else               next_if = cpu_if;
          else                 next_if = cpu_if;
        end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a|jmp_a)      next_if = add[15:1] + 15'h0002;
          else if(jmp_i)
            case(cpu_a[1:0])
              0 :              next_if = cpu_dar0[15:1] + 15'h0002;
              1 :              next_if = cpu_dar1[15:1] + 15'h0002;
              2 :              next_if = cpu_dar2[15:1] + 15'h0002;
              3 :              next_if = cpu_dar3[15:1] + 15'h0002;
            endcase
          else                 next_if = cpu_if;
        end
//****************************************************************************
      S_IND_Y,S_REG_OP,S_BCD_DELAY:
        begin
          if(iCount < 4'h9)    next_if = cpu_ifp4;
          else                 next_if = cpu_if;
        end
//****************************************************************************
      S_RTS:                   next_if = pc_p1r[15:1] + 15'h0002;
//****************************************************************************
      S_IR_W_HIT:              next_if = pc[15:1] + 15'h0002;
//****************************************************************************
      default:                 next_if = cpu_if;
    endcase
end
 
//****************************************************************************
//****************************************************************************
//next if look-ahead
always@(*)
  begin
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(brk)          next_sp = spm3;
          else if(rts)     next_sp = spp2;
          else if(rti)     next_sp = spp3;
          else if(txs)     next_sp = x;
          else if(plp|pla) next_sp = spp1;
          else if(pha|php) next_sp = spm1;
          else             next_sp = sp;
        end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a)        next_sp = spm2;
          else             next_sp = sp;
       end
//****************************************************************************
      S_NMI,S_IRQ:         next_sp = spm3;
//****************************************************************************
      default:             next_sp = sp;
    endcase
  end
  
//****************************************************************************
//****************************************************************************
//next C flag
always@(*)
  begin
    case(state)
//****************************************************************************
//start executing a single byte instruction
      S_EXECUTE_INST1:
        begin
          if(brk|rts) next_C = carry;
          else if(rti)
            begin
              case(spp1[1:0])
                0: next_C = caches[0];
                1: next_C = caches[8];
                2: next_C = caches[16];
                3: next_C = caches[24];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0: next_C = caches[8];
                    1: next_C = caches[16];
                    2: next_C = caches[24];
                    3: next_C = caches[0];
                  endcase
                end
              else if(asla | rola) next_C = a[7];
              else if(lsra | rora) next_C = a[0];
              else if(clc)         next_C = FALSE;
              else if(sec)         next_C = TRUE;
              else                 next_C = carry;
            end
        end
//****************************************************************************
//start executing a 2 byte instruction
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(imm_addsub & ~decimal) next_C = Cout_a;
              else if(cmpa_i)           next_C = Cout_ca;
              else if(cpy_i)            next_C = Cout_y;
              else if(cpx_i)            next_C = Cout_x;
              else                      next_C = carry;
            end
          else next_C = carry;
        end
//****************************************************************************
      S_WR_MEM_OP: // complete memory operation
        begin
          if(asl | rol)      next_C = alu_op[7];
          else if(lsr | ror) next_C = alu_op[0];
          else               next_C = carry;
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(cmpa_2)                   next_C = Cout_ca;
          else if(cpy_2)               next_C = Cout_y;
          else if(cpx_2)               next_C = Cout_x;
          else if(addsub_2 & ~decimal) next_C = Cout_a;
          else                         next_C = carry;
        end
//****************************************************************************
      S_BCD_DELAY:                     next_C = bcdCout_a;
//****************************************************************************
      default:                         next_C = carry;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next V flag
always@(*)
  begin
    case(state)
//****************************************************************************
//start executing a single byte instruction
      S_EXECUTE_INST1:
        begin
          if(brk|rts) next_V = overflow;
          else if(rti)
            begin
              case(spp1[1:0])
                0: next_V = caches[6];
                1: next_V = caches[14];
                2: next_V = caches[22];
                3: next_V = caches[30];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0: next_V = caches[14];
                    1: next_V = caches[22];
                    2: next_V = caches[30];
                    3: next_V = caches[6];
                  endcase
                end
              else if(clv) next_V = FALSE;
              else         next_V = overflow;
            end
        end
//****************************************************************************
//start executing a 2 byte instruction
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(imm_addsub & ~decimal) next_V = is_overflow_a;
              else                      next_V = overflow;
            end
          else                          next_V = overflow;
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(bita)                     next_V = alu_op[6];
          else if(addsub_2 & ~decimal) next_V = is_overflow_a;
          else                         next_V = overflow;
        end
//****************************************************************************
      S_BCD_DELAY:                     next_V = bcdOverflow_a;
//****************************************************************************
      default:                         next_V = overflow;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next N flag
always@(*)
  begin
    case(state)
//****************************************************************************
//start executing a single byte instruction
      S_EXECUTE_INST1:
        begin
          if(brk|rts) next_N = negative;
          else if(rti)
            begin
              case(spp1[1:0])
                0: next_N = caches[7];
                1: next_N = caches[15];
                2: next_N = caches[23];
                3: next_N = caches[31];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0: next_N = caches[15];
                    1: next_N = caches[23];
                    2: next_N = caches[31];
                    3: next_N = caches[7];
                  endcase
                end
              else if(asla | rola) next_N = a[6];
              else if(lsra)        next_N = L;
              else if(rora)        next_N = carry;
              else if(tya)         next_N = y[7];
              else if(txa)         next_N = x[7];
              else if(tay | tax)   next_N = a[7];
              else if(dey)         next_N = (y == ZERO) ? H : (y == 8'h80) ? L : y[7];
              else if(iny)         next_N = (y == NEG_ONE) ? L : (y == 8'h7F) ? H : y[7];
              else if(tsx)         next_N = sp[7];
              else if(dex)         next_N = (x == ZERO) ? H : (x == 8'h80) ? L : x[7];
              else if(inx)         next_N = (x == NEG_ONE) ? L : (x == 8'h7F) ? H : x[7];
              else                 next_N = negative;
            end
        end
//****************************************************************************
//start executing a 2 byte instruction
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(anda_i)                     next_N = and_op[7];
              else if(eora_i)                next_N = eor_op[7];
              else if(ora_i)                 next_N = or_op[7];
              else if(lda_i | ldy_i | ldx_i) next_N = alu_op[7];
              else if(imm_addsub & ~decimal) next_N = is_negative_a;
              else if(cmpa_i)                next_N = N_ca;
              else if(cpy_i)                 next_N = N_y;
              else if(cpx_i)                 next_N = N_x;
              else                           next_N = negative;
            end
          else                               next_N = negative;
        end
//****************************************************************************
      S_WR_MEM_OP: // complete memory operation
        begin
          if(asl | rol) next_N = alu_op[6];
          else if(lsr)  next_N = L;
          else if(ror)  next_N = carry;
          else if(dec)  next_N = alu_op_dec_neg;
          else          next_N = alu_op_inc_neg;
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(ora)                         next_N = or_op[7];
          else if(bita | ldx | lda | ldy) next_N = alu_op[7];
          else if(anda)                   next_N = and_op[7];
          else if(eora)                   next_N = eor_op[7];
          else if(cmpa_2)                 next_N = N_ca;
          else if(cpy_2)                  next_N = N_y;
          else if(cpx_2)                  next_N = N_x;
          else if(addsub_2 & ~decimal)    next_N = is_negative_a;
          else                            next_N = negative;
        end
//****************************************************************************
      S_BCD_DELAY:                        next_N = bcdNegative_a;
//****************************************************************************
      default:                            next_N = negative;
    endcase
  end

//****************************************************************************
//****************************************************************************
//next Z flag
always@(*)
  begin
    case(state)
//****************************************************************************
//start executing a single byte instruction
      S_EXECUTE_INST1:
        begin
          if(brk|rts) next_Z = zero;
          else if(rti)
            begin
              case(spp1[1:0])
                0: next_Z = caches[1];
                1: next_Z = caches[9];
                2: next_Z = caches[17];
                3: next_Z = caches[25];
              endcase
            end
//****************************************************************************
// 1 byte instructions
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0: next_Z = caches[9];
                    1: next_Z = caches[17];
                    2: next_Z = caches[25];
                    3: next_Z = caches[1];
                  endcase
                end
              else if(asla) next_Z = ~(|a[6:0]);
              else if(rola) next_Z = ~carry & ~(|a[6:0]);
              else if(lsra) next_Z = ~(|a[7:1]);
              else if(rora) next_Z = ~carry & ~(|a[7:1]);
              else if(tya)  next_Z = (y == ZERO);
              else if(txa)  next_Z = (x == ZERO);
              else if(tay)  next_Z = (a == ZERO);
              else if(dey)  next_Z = (y == ONE);
              else if(iny)  next_Z = (y == NEG_ONE);
              else if(tax)  next_Z = (a == ZERO);
              else if(tsx)  next_Z = (sp == ZERO);
              else if(dex)  next_Z = (x == ONE);
              else if(inx)  next_Z = (x == NEG_ONE);
              else          next_Z = zero;
            end
        end
//****************************************************************************
//start executing a 2 byte instruction
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(anda_i)                     next_Z = and_op_zero;
              else if(eora_i)                next_Z = eor_op_zero;
              else if(ora_i)                 next_Z = or_op_zero;
              else if(lda_i | ldy_i | ldx_i) next_Z = alu_op_0;
              else if(imm_addsub & ~decimal) next_Z = is_zero_a;
              else if(cmpa_i)                next_Z = is_zero_ca;
              else if(cpy_i)                 next_Z = is_zero_y;
              else if(cpx_i)                 next_Z = is_zero_x;
              else                           next_Z = zero;
            end
          else                               next_Z = zero;
        end
//****************************************************************************
      S_WR_MEM_OP: // complete memory operation
        begin
          if(asl)                      next_Z = ~(|alu_op[6:0]);
          else if(rol)                 next_Z = ~carry & ~(|alu_op[6:0]);
          else if(lsr)                 next_Z = ~(|alu_op[7:1]);
          else if(ror)                 next_Z = ~carry & ~(|alu_op[7:1]);
          else if(dec)                 next_Z = alu_op_1;
          else                         next_Z = alu_op_255;
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(ora)                      next_Z = or_op_zero;
          else if(bita | anda)         next_Z = and_op_zero;
          else if(eora)                next_Z = eor_op_zero;
          else if(cmpa_2)              next_Z = is_zero_ca;
          else if(cpy_2)               next_Z = is_zero_y;
          else if(cpx_2)               next_Z = is_zero_x;
          else if(ldx | lda | ldy)     next_Z = alu_op_0;
          else if(addsub_2 & ~decimal) next_Z = is_zero_a;
          else                         next_Z = zero;
        end
//****************************************************************************
      S_BCD_DELAY:                     next_Z = bcdZero_a;
//****************************************************************************
      default:                         next_Z = zero;
    endcase
  end


//synchronous cpu state machine
//****************************************************************************
//****************************************************************************
//instruction execution state machine controller
always@(posedge clk)
  begin
    //bus master logic
    vic_ack <= vic_req;
    dma_ack <= ~vic_req|dma_req|~dma_ack; //must only be a one cycle steal
    cpu_ack <= ~(vic_req & ~(vic_req & dma_ack & ~dma_req) & cpu_pause & resetIn & (&halt));    
       
    //cpu is being reset
    if(~resetIn)
      begin

`ifdef debug
        pc_trace <= 960'd0;
        ad_trace <= 960'd0;
        aw_trace <= 960'd0;
        breakTrig <= FALSE;
        breakTrig1 <= FALSE;
        cycleCount <= 32'h00000000; //count events
        cycleCount2 <= 32'h00000000; //count events
        enableCycleCount2 <= H;
`endif
        
        halt <= 4'hF; 

        rnmi <= H;
        rnmi2 <= H; //this needs to be cleared to allow nmi's
        nmi <= H;
        rirq <= H;
        
        state <= S_RESET;

//the rest of these are optional
        iCount <= 0;
        stackit <= 0;
      end
    else if(~cpu_ack) //cpu is the master
      begin
      
        state <= next_state;
        
        //a <= next_a;
        x <= next_x;
        y <= next_y;
        
        carry <= next_C;
        overflow <= next_V;
        negative <= next_N;
        zero <= next_Z;

        addx <= { ah, imm_op } + next_x;
        addy <= { ah, imm_op } + next_y;
        
        pc <= next_pc;
        pc_p1r <= pcp1;
        pc_p2r <= pcp2;
        pc_p3r <= pcp3;
        pc_p4r <= pcp4;
        pc_p5r <= pcp5;

        cpu_if <= next_if;
        
        sp <= next_sp;
        spp1 <= next_spp1;
        spp2 <= next_spp2;
        spp3 <= next_spp3;
        
        if(state == S_EXECUTE_INST1)      branch <= pc_p3r + { {9{imm_op[7]}}, imm_op[6:0] };
        else if(state == S_EXECUTE_INST2) branch <= pc_p4r + { {9{imm_op[7]}}, imm_op[6:0] };
        else if(state == S_EXECUTE_INST3) branch <= pc_p5r + { {9{imm_op[7]}}, imm_op[6:0] };
        //not in an execute state, ir and pc must be valid for the next instruction
        else                              branch <= pc_p2r + { {9{imm_op[7]}}, imm_op[6:0] };
        
        if((next_state == S_EXECUTE_INST2) |
           (next_state == S_EXECUTE_INST3))     mode <= ira[7];
        
        //load #, zp cache operand or memory operand
        if(next_state == S_EXECUTE_INST2)       alu_op <= imm_op;
        else if(next_state == S_BCD_DELAY)      alu_op <= alu_op;
        else if((state == S_EXECUTE_INST1) |
                (state == S_EXECUTE_INST2) |
                ((state == S_EXECUTE_INST3) & cacherw_hit))
          begin
            case(cpu_aw[1:0])
              0:                                alu_op <= cache[7:0];
              1:                                alu_op <= cache[15:8];
              2:                                alu_op <= cache[23:16];
              3:                                alu_op <= cache[31:24];
            endcase
          end
        else if(pRead)                          alu_op <= cpu_db;
        else
          case(cpu_a[1:0])
            0 :                                 alu_op <= cpu_rd[7:0];
            1 :                                 alu_op <= cpu_rd[15:8];
            2 :                                 alu_op <= cpu_rd[23:16];
            3 :                                 alu_op <= cpu_rd[31:24];
          endcase
   
`ifdef debug

        ad_trace[15:0] <= cpu_a;
        ad_trace[959:16] <= ad_trace[943:0];
        aw_trace[15:0] <= cpu_aw;
        aw_trace[959:16] <= aw_trace[943:0];

        breakTrig <= breakTrig1;

        cpu_a2_i <= cpu_a;
        cpu_al_i <= cpu_al;
        cpu_rw_r <= cpu_rw;
        cpu_byte_enable <= cpu_wbyte;
        cpu_drr <= cpu_rd;

        cpu_aw2_i <= cpu_aw;
        cpu_rwz_r <= cpu_rwz;
        cpu_byte_enablez <= cpu_wbytez;
        cache_rr <= cache;

        cpu_pr_r <= cpu_db;
        cpu_wb_r <= cpu_wb;

        //count events
        //if(state == S_RESET) cycleCount <= 0; //count events
        //else cycleCount <= cycleCount + H;
        
        if(breakTrig) enableCycleCount2 <= L;
`endif
        
        //update interrupts
        rnmi <= nmi_req;
        rnmi2 <= rnmi;
        rirq <= irq | interrupt;
        
        if(~rnmi & rnmi2) nmi <= L; //edge trigger until processed
        
        case(state)
//****************************************************************************
          S_RESET:  //accessing reset vector (pc) from memory
            begin
              tmp_a <= V_RST;//reset vector
            end
//****************************************************************************
//transfer the first instruction from the new path to cpu_dr & start accessing
//the next 4 bytes in the instruction stream
          S_LOAD_INST:
            begin
              case(pc[1:0])
                2'h0 :
                  begin
                    ir[95:32] <= 0;
                    ir[31:0] <= cpu_rd[31:0];
                    iCount <= 4'h4;
                  end
                2'h1 :
                  begin
                    ir[95:24] <= 0;
                    ir[23:0] <= cpu_rd[31:8];
                    iCount <= 4'h3;
                  end
                2'h2 :
                  begin
                    ir[95:32] <= 0;
                    ir[15:0] <= cpu_rd[31:16];
                    ir[31:16] <= cpu_rd[15:0];
                    iCount <= 4'h4;
                  end
                2'h3 :
                  begin
                    ir[95:24] <= 0;
                    ir[7:0] <= cpu_rd[31:24];
                    ir[23:8] <= cpu_rd[15:0];
                    iCount <= 4'h3;
                  end
              endcase
            end

//****************************************************************************
//transfer the next instruction bytes from the new path to cpu_rd & start accessing
//the next 4 bytes in the instruction stream
          S_DECODE_INST:
            begin
              case(iCount[0])
                L :
                  begin
                    ir[63:32] <= ib32;
                  end
                H :
                  begin
                    ir[55:24] <= ib32;
                  end
              endcase
              iCount <= iCount + 4'h4;
            end
            
//****************************************************************************
//start executing a single byte instruction
          S_EXECUTE_INST1:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif

//remember operations for later
              sta_iy_2 <= sta_iy;
              
              asl <= asl_z|asl_zx|asl_a|asl_ax;
              rol <= rol_z|rol_zx|rol_a|rol_ax;
              lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
              ror <= ror_z|ror_zx|ror_a|ror_ax;
              dec <= dec_z|dec_zx|dec_a|dec_ax;
              inc <= inc_z|inc_zx|inc_a|inc_ax;

              ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
              bita <= bita_z|bita_a;
              anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
              eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
              addsub_2 <= addsub;
              
              cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
              cpx_2 <= cpx_z|cpx_a;
              cpy_2 <= cpy_z|cpy_a;
              
              ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
              lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
              ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;

              mem_op_2 <= mem_op;
              
              //need to remember this in case an interupt is occuring during this instruction
              jsr_a_2 <= jsr_a;

              //stacking stuff for an interrupt, brk or jsr
              stackit <= 0;

//deal with program counter change possibilities first...
              if(brk)
                begin
                  stackit <= { H, pc[7:0], sp_brk };
//brk is same as irq except for break status bit set on the stack and pc + 2
//is stacked, note this is a software interrupt, uses irq vector
                  interrupt <= H;
                  tmp_a <= V_IRQ;  
                end
              else if(rts)
                begin
                  stackit <= 1;
                end
              else if(rti)
                begin
                  case(spp1[1:0])
                    0:
                      begin
                        interrupt <= caches[2];
                        decimal <= caches[3];
                      end
                    1:
                      begin
                        interrupt <= caches[10];
                        decimal <= caches[11];
                      end
                    2:
                      begin
                        interrupt <= caches[18];
                        decimal <= caches[19];
                      end
                    3:
                      begin
                        interrupt <= caches[26];
                        decimal <= caches[27];
                     end
                  endcase
                end
              else
                begin
                  if(iCount < 4'hA)
                    begin
                      //subtracting 1 and adding 4
                      iCount <= iCount + 4'h3;
                      case(iCount)
                      
                        4'h9:
                          begin
                            ir[63:0] <= ir[71:8];
                            ir[95:64] <= ib32;
                          end
                        4'h8:
                          begin
                            ir[55:0] <= ir[63:8];
                            ir[87:56] <= ib32;
                          end
                        4'h7:
                          begin
                            ir[47:0] <= ir[55:8];
                            ir[79:48] <= ib32;
                          end
                        4'h6:
                          begin
                            ir[39:0] <= ir[47:8];
                            ir[71:40] <= ib32;
                          end
                        4'h5:
                          begin
                            ir[31:0] <= ir[39:8];
                            ir[63:32] <= ib32;
                          end
                        4'h4:
                          begin
                            ir[23:0] <= ir[31:8];
                            ir[55:24] <= ib32;
                          end
                        4'h3:
                          begin
                            ir[15:0] <= ir[23:8];
                            ir[47:16] <= ib32;
                          end
                        4'h2:
                          begin
                            ir[7:0] <= ir[15:8];
                            ir[39:8] <= ib32;
                          end
                        4'h1:
                          begin
                            ir[31:0] <= ib32;
                          end

                        //this should not happen... 
                        default: halt <= 4'h0; 
                      endcase
                    end
                  else
                    begin
                      iCount <= iCount - 4'h1;
                      ir[95:88] <= ZERO;
                      ir[87:0] <= ir[95:8];
                    end
                
                  if(plp)
                    begin
                      case(sp[1:0])
                        0:
                          begin
                            interrupt <= caches[10];
                            decimal <= caches[11];
                          end
                        1:
                          begin
                            interrupt <= caches[18];
                            decimal <= caches[19];
                          end
                        2:
                          begin
                            interrupt <= caches[26];
                            decimal <= caches[27];
                          end
                        3:
                          begin
                            interrupt <= caches[2];
                            decimal <= caches[3];
                          end
                      endcase
                    end

                  if(pla)
                    begin
                      tmp_a <= spp1;
                      case(sp[1:0])
                        0: alu_op <= caches[15:8];
                        1: alu_op <= caches[23:16];
                        2: alu_op <= caches[31:24];
                        3: alu_op <= caches[7:0];
                      endcase
                    end
                    
                  if(asla) a <= { a[6:0], L };
                  if(rola) a <= { a[6:0], carry };
                  if(lsra) a <= { L, a[7:1] };
                  if(rora) a <= { carry, a[7:1] };
                  if(tya) a <= y;
                  if(txa) a <= x;

                  if(cli) interrupt <= FALSE;
                  if(sei) interrupt <= TRUE;
                  if(cld) decimal <= FALSE;
                  if(sed) decimal <= TRUE;
                end
            end
//****************************************************************************
//start executing a 2 byte instruction
          S_EXECUTE_INST2:
            begin

`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
              if(enableCycleCount2 & (st_zxy | mem_op)) cycleCount2 <= cycleCount2 + H;
`endif

//remember operations for later
              sta_iy_2 <= sta_iy;
              
              asl <= asl_z|asl_zx|asl_a|asl_ax;
              rol <= rol_z|rol_zx|rol_a|rol_ax;
              lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
              ror <= ror_z|ror_zx|ror_a|ror_ax;
              dec <= dec_z|dec_zx|dec_a|dec_ax;
              inc <= inc_z|inc_zx|inc_a|inc_ax;

              ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
              bita <= bita_z|bita_a;
              anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
              eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
              addsub_2 <= addsub;
              
              cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
              cpx_2 <= cpx_z|cpx_a;
              cpy_2 <= cpy_z|cpy_a;
              
              ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
              lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
              ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;

              mem_op_2 <= mem_op;
              
              //need to remember this in case an interupt is occuring during this instruction
              jsr_a_2 <= jsr_a;

              //stacking stuff for an interrupt, brk or jsr
              stackit <= 0;

//deal with program counter change possibilities first...
              if(branch_taken)
                begin
                  stackit <= 0; //nothing to be done here 
                  case(branch[1:0])
                    2'h0 :
                      begin
                        ir[95:32] <= 0;
                        ir[31:0] <= cpu_rd[31:0];
                        iCount <= 4'h4;
                      end
                    2'h1 :
                      begin
                        ir[95:24] <= 0;
                        ir[23:0] <= cpu_rd[31:8];
                        iCount <= 4'h3;
                      end
                    2'h2 :
                      begin
                        ir[95:32] <= 0;
                        ir[15:0] <= cpu_rd[31:16];
                        ir[31:16] <= cpu_rd[15:0];
                        iCount <= 4'h4;
                      end
                    2'h3 :
                      begin
                        ir[95:24] <= 0;
                        ir[7:0] <= cpu_rd[31:24];
                        ir[23:8] <= cpu_rd[15:0];
                        iCount <= 4'h3;
                      end
                  endcase
                end
              else if(two_byte) //necessary to catch illegal instructions below
                begin
                  if(~st_zxy)
                    begin
                      //subtracting 2 and adding 4
                      iCount <= iCount + 4'h2;
                      case(iCount)

                        4'hA:
                          begin
                            ir[63:0] <= ir[79:16];
                            ir[95:64] <= ib32;
                          end
                        4'h9:
                          begin
                            ir[55:0] <= ir[71:16];
                            ir[87:56] <= ib32;
                          end
                        4'h8:
                          begin
                            ir[47:0] <= ir[63:16];
                            ir[79:48] <= ib32;
                          end
                        4'h7:
                          begin
                            ir[39:0] <= ir[55:16];
                            ir[71:40] <= ib32;
                          end
                        4'h6:
                          begin
                            ir[31:0] <= ir[47:16];
                            ir[63:32] <= ib32;
                          end
                        4'h5:
                          begin
                            ir[23:0] <= ir[39:16];
                            ir[55:24] <= ib32;
                          end
                        4'h4:
                          begin
                            ir[15:0] <= ir[31:16];
                            ir[47:16] <= ib32;
                          end
                        4'h3:
                          begin
                            ir[7:0] <= ir[23:16];
                            ir[39:8]  <= ib32;
                          end

                        4'h2:
                          begin
                            ir[31:0]  <= ib32;
                          end

                        default:
                          begin
                            //halt <= 4'h1; //this should not happen
                            iCount <= iCount - 4'h2;
                            ir[95:80] <= 16'h0000;
                            ir[79:0] <= ir[95:16];
                          end
                      endcase
                    end
                  else
                    begin
                      iCount <= iCount - 4'h2;
                      ir[95:80] <= 16'h0000;
                      ir[79:0] <= ir[95:16];
                    end

                  if(anda_i) a <= and_op;
                  if(eora_i) a <= eor_op;
                  if(ora_i) a <= or_op;
                  if(lda_i) a <= alu_op;
                  if(imm_addsub & ~decimal) a <= result;
                  
                  if(ldx_zy|stx_zy) tmp_a <= { ZERO, addy[7:0] };
                  else if(zx) tmp_a <= { ZERO, addx[7:0] };
                  else if(ix)
                    begin
                      case(addx[1:0])
                        2'h0 : tmp_a <= cpu_zar0;
                        2'h1 : tmp_a <= cpu_zar1;
                        2'h2 : tmp_a <= cpu_zar2;
                        2'h3 : tmp_a <= cpu_zar3;
                      endcase
                    end
                  else if(iy)  //cacheiSelect <= add[1:0];
                    begin
                      case(add[1:0])
                        2'h0 : tmp_a <= cpu_zar0;
                        2'h1 : tmp_a <= cpu_zar1;
                        2'h2 : tmp_a <= cpu_zar2;
                        2'h3 : tmp_a <= cpu_zar3;
                      endcase
                    end
                  else         tmp_a <= { ZERO, add[7:0] };
                end
//****************************************************************************
//illegal instructions
              else halt <= 4'h2;
            end
//****************************************************************************
//start executing a 3 byte instruction
          S_EXECUTE_INST3:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
              if(enableCycleCount2 & (st_axy | mem_op)) cycleCount2 <= cycleCount2 + H;
`endif

//remember operations for later
              sta_iy_2 <= sta_iy;
              
              asl <= asl_z|asl_zx|asl_a|asl_ax;
              rol <= rol_z|rol_zx|rol_a|rol_ax;
              lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
              ror <= ror_z|ror_zx|ror_a|ror_ax;
              dec <= dec_z|dec_zx|dec_a|dec_ax;
              inc <= inc_z|inc_zx|inc_a|inc_ax;

              ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
              bita <= bita_z|bita_a;
              anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
              eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
              addsub_2 <= addsub;
              
              cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
              cpx_2 <= cpx_z|cpx_a;
              cpy_2 <= cpy_z|cpy_a;
              
              ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
              lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
              ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;

              mem_op_2 <= mem_op;
              
              //need to remember this in case an interupt is occuring during this instruction
              jsr_a_2 <= jsr_a;
              
              //stacking stuff for an interrupt, brk or jsr
              stackit <= 0;

//deal with program counter change possibilities first...
              if(jsr_a)
                begin
                  stackit <= { H, pc_p2r };
                  case(add[1:0])
                    2'h0 :
                      begin
                        ir[95:32] <= 0;
                        ir[31:0] <= cpu_rd[31:0];
                        iCount <= 4'h4;
                      end
                    2'h1 :
                      begin
                        ir[95:24] <= 0;
                        ir[23:0] <= cpu_rd[31:8];
                        iCount <= 4'h3;
                      end
                    2'h2 :
                      begin
                        ir[95:32] <= 0;
                        ir[15:0] <= cpu_rd[31:16];
                        ir[31:16] <= cpu_rd[15:0];
                        iCount <= 4'h4;
                      end
                    2'h3 :
                      begin
                        ir[95:24] <= 0;
                        ir[7:0] <= cpu_rd[31:24];
                        ir[23:8] <= cpu_rd[15:0];
                        iCount <= 4'h3;
                      end
                  endcase
                end                  
              else if(jmp_a)
                begin
                  stackit <= 1; //nothing to be done here
                  case(add[1:0])
                    2'h0 :
                      begin
                        ir[95:32] <= 0;
                        ir[31:0] <= cpu_rd[31:0];
                        iCount <= 4'h4;
                      end
                    2'h1 :
                      begin
                        ir[95:24] <= 0;
                        ir[23:0] <= cpu_rd[31:8];
                        iCount <= 4'h3;
                      end
                    2'h2 :
                      begin
                        ir[95:32] <= 0;
                        ir[15:0] <= cpu_rd[31:16];
                        ir[31:16] <= cpu_rd[15:0];
                        iCount <= 4'h4;
                      end
                    2'h3 :
                      begin
                        ir[95:24] <= 0;
                        ir[7:0] <= cpu_rd[31:24];
                        ir[23:8] <= cpu_rd[15:0];
                        iCount <= 4'h3;
                      end
                  endcase
                end
              else if(jmp_i)
                begin
                  tmp_a <= add;
                end
              else
                begin
                  iCount <= iCount - 4'h3;
                  ir[95:72] <= 24'h000000;
                  ir[71:0] <= ir[95:24];

                  if(ax)      tmp_a <= addx;
                  else if(ay) tmp_a <= addy;
                  else        tmp_a <= add;
                end
            end
//****************************************************************************
          S_IND_Y:  // indirect y 
            begin
              tmp_a <= tmp_a + y;
              if(iCount < 4'h9)
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h5; //this should not happen
                  endcase
                end
            end
//****************************************************************************
          S_REG_OP: // complete register operation
            begin
              if(iCount < 4'h9)
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'h9; //this should not happen
                  endcase
                end

              if(ora) a <= or_op;
              if(anda) a <= and_op;
              if(eora) a <= eor_op;
              if(lda) a <= alu_op;
              if(addsub_2 & ~decimal) a <= result;
            end
//****************************************************************************
          S_BCD_DELAY: //bcd arithmatic delay
            begin
              if(iCount < 4'h9)
                begin
                  iCount <= iCount + 4'h4;
                  case(iCount)
                    4'h8: ir[95:64] <= ib32;
                    4'h7: ir[87:56] <= ib32;
                    4'h6: ir[79:48] <= ib32;
                    4'h5: ir[71:40] <= ib32;
                    4'h4: ir[63:32] <= ib32;
                    4'h3: ir[55:24] <= ib32;
                    4'h2: ir[47:16] <= ib32;
                    4'h1: ir[39:8]  <= ib32;
                    4'h0: ir[31:0]  <= ib32;
                    default: halt <= 4'hA; //this should not happen
                  endcase
                end

              a <= bcdResult;
            end
//****************************************************************************
//start processing an nmi
          S_NMI:
            begin
              stackit <= { H, pc[7:0], sp_int };
//re-enable NMI, cannot become active until the first instruction of service routine
              nmi <= TRUE; 
              tmp_a <= V_NMI;
            end
//****************************************************************************
//start processing an irq
          S_IRQ:
            begin
              stackit <= { H, pc[7:0], sp_int };
              interrupt <= H;
              tmp_a <= V_IRQ;
            end
//****************************************************************************
//reload ir register
          S_IR_W_HIT:
            begin
              stackit <= 0;
              case(pc[1:0])
                2'h0 :
                  begin
                    ir[95:32] <= 0;
                    ir[31:0] <= cpu_rd[31:0];
                    iCount <= 4'h4;
                  end
                2'h1 :
                  begin
                    ir[95:24] <= 0;
                    ir[23:0] <= cpu_rd[31:8];
                    iCount <= 4'h3;
                  end
                2'h2 :
                  begin
                    ir[95:32] <= 0;
                    ir[15:0] <= cpu_rd[31:16];
                    ir[31:16] <= cpu_rd[15:0];
                    iCount <= 4'h4;
                  end
                2'h3 :
                  begin
                    ir[95:24] <= 0;
                    ir[7:0] <= cpu_rd[31:24];
                    ir[23:8] <= cpu_rd[15:0];
                    iCount <= 4'h3;
                  end
              endcase
`ifdef debug
              cycleCount[7:0] <= cycleCount[7:0] + H;
`endif
            end
//****************************************************************************
          S_IR_LOAD:
            begin
              iCount <= iCount + 4'h4;
              case(iCount)
                4'h8: ir[95:64] <= ib32;
                4'h7: ir[87:56] <= ib32;
                4'h6: ir[79:48] <= ib32;
                4'h5: ir[71:40] <= ib32;
                4'h4: ir[63:32] <= ib32;
                4'h3: ir[55:24] <= ib32;
                4'h2: ir[47:16] <= ib32;
                4'h1: ir[39:8]  <= ib32;
                4'h0: ir[31:0]  <= ib32;
                default: halt <= 4'h5; //this should not happen
              endcase
`ifdef debug
              cycleCount[31:8] <= cycleCount[31:8] + H;
`endif
            end
        endcase
    end
end

`ifdef debug

cpu_debug dbg(clk, pc_trace, ad_trace, aw_trace
             ,state,halt,cpu_rwz_r,cpu_byte_enablez,cache_rr,cpu_rw_r,cpu_byte_enable,a1
             ,negative,decimal,interrupt,overflow,zero,carry,rirq,nmi
             ,ir,iCount
             ,a,x,y,sp,alu_op,cpu_pr_r,cpu_wb_r
             ,pc,tmp_a,cpu_if,cpu_a2_i,cpu_al_i,cpu_aw2_i,break_pc
             ,cpu_drr, cycleCount, cycleCount2
             ,debug_add, debug_Dout, debug_sel
             );
`endif

endmodule


